"""Body assertions module."""
from typing import Optional

from ..base_request import BaseRequest
from ...decorators import then

from truth.truth import AssertThat
from pydash import _


class BodyAssertions(BaseRequest):
    """Content-Type: Application/json responses assertions."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    @then
    def assert_response_contains(self, expected_key):
        """Assert response body contains key."""
        is_present = _.get(self.response.json(), expected_key)
        AssertThat(is_present).IsTruthy()

    @then
    def assert_response_field(self, target_key):
        """Set a target value to assert any other condition."""
        self.target = _.get(self.response.json(), target_key)

    @then
    def is_empty(self):
        """Assert target value is an empty array."""
        AssertThat(self.target).IsEmpty()

    @then
    def is_not_empty(self):
        """Assert target value is an empty array."""
        AssertThat(self.target).IsNotEmpty()

    @then
    def is_true(self):
        """Assert target value is true."""
        AssertThat(self.target).IsTrue()

    @then
    def is_false(self):
        """Assert target value is false."""
        AssertThat(self.target).IsFalse()

    @then
    def is_null(self):
        """Assert target value is null."""
        AssertThat(self.target).IsNone()

    @then
    def equals(self, expected_value):
        """Assert target value is null."""
        AssertThat(self.target).IsEqualTo(expected_value)

    @then
    def contains(self, expected_value):
        """Assert target value is null."""
        AssertThat(self.target).Contains(expected_value)

    @then
    def does_not_contain(self, expected_value):
        """Assert target value is null."""
        AssertThat(self.target).DoesNotContain(expected_value)

    @then
    def is_equal_to_request_payload_field(self, key: str):
        """Asserts target value equals to Request().body key."""
        AssertThat(self.target).IsEqualTo(_.get(self.body, key))

    @then
    def is_not_equal_to_request_payload_field(self, key: str):
        """Asserts target value is not equal to Request().body key."""
        AssertThat(self.target).IsNotEqualTo(_.get(self.body, key))



"""Headers assertions module."""
from typing import Optional

from ..base_request import BaseRequest
from ...decorators import then

from truth.truth import AssertThat
from pydash import _


class HeadersAssertions(BaseRequest):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    @then
    def assert_response_headers_contains(self, key: str):
        AssertThat(self.response.headers).Contains(key)

    @then
    def assert_response_headers_does_not_contain(self, key: str):
        AssertThat(self.response.headers).Contains(key)

    @then
    def assert_response_header(self, key: str):
        self._target = self.response.headers.get(key)

    @then
    def equals(self, value):
        AssertThat(self._target).IsEqualTo(value)

    @then
    def is_not_equal_to(self, value):
        AssertThat(self._target).IsNotEqualTo(value)



"""Aliases and exports."""
from .action_request import Actions
from .precondition_request import Preconditions
from .assertions import BodyAssertions, HeadersAssertions, StatusAssertions


class __Request(
    Actions, BodyAssertions, HeadersAssertions, Preconditions, StatusAssertions
):
    ...


Request = __Request
AssertioRequest = __Request

__all__ = ("AssertioRequest", "Request")



"""Config functions."""
import os
import json

from dataclasses import dataclass, field
from pathlib import Path

import yaml


def _stat(filename: str) -> Path:
    """Return a file if it exists."""
    file = Path.cwd().joinpath(filename)
    if file.exists():
        return file

@dataclass
class Config:
    """Configuration namespace."""

    base_url: str = field(default=os.getenv("ASSERTIO_BASE_URL", ""))
    logfile: str = field(default=os.getenv("ASSERTIO_LOGFILE", "assertio.log"))
    payloads_dir: str = field(
        default=os.getenv("ASSERTIO_PAYLOADS_DIR", "features/payloads")
    )

    def from_json(self, config_file: str = "assertio.json"):
        """Create config object from a json file."""
        file = _stat(config_file)
        if file is not None:
            config_json = json.load(open(file))
            for key, value in config_json.items():
                setattr(self, key, value)

    def from_yaml(self, config_file: str = "assertio.yaml"):
        """Create config object from a yaml file."""
        file = _stat(config_file)
        if file is not None:
            config_yaml = yaml.safe_load(open(file))
            for key, value in config_yaml.items():
                setattr(self, key, value)



from argparse import ArgumentParser
from pathlib import Path
from pydoc import importfile

from .config import Config


JSON_NAME = "assertio.json"
YAML_NAME = "assertio.yaml"


class _CLI:
    def __init__(self):
        _ = ArgumentParser()
        _.add_argument(
            "--run",
            default="all",
        )
        _.add_argument(
            "--settings",
            "-s",
            default=None
        )
        self._args = _.parse_args()

    def load_modules(self):
        modules = []
        for p in Path.cwd().joinpath("features/runners").glob("**/*.py"):
            modules.append(importfile(str(p)))
        return modules
    
    def get_runners_for(self, module):
        return [
            getattr(module, name) 
            for name in dir(module) 
            if name.endswith("Runner") and name != "Runner"
        ]

    def run_all(self):
        for module in self.load_modules():
            for runner in self.get_runners_for(module):
                runner().start()
    
    def run_one(self):
        for module in self.load_modules():
            for runner in self.get_runners_for(module):
                if self._args.run == runner.__name__:
                    runner().start()

    def get_config(self) -> Config:
        _ = Config()
        if self._args.settings is None:
            _.from_json(JSON_NAME)
            _.from_yaml(YAML_NAME)
        elif self._args.settings.endswith("json"):
            _.from_json(self._args.settings)
        elif self._args.settings.endswith("yaml"):
            _.from_yaml(self._args.settings)
        
        return _


    def bootstrap(self):
        if self._args.run == "all":
            self.run_all()
        else:
            self.run_one()



import json

from pathlib import Path
from typing import Dict, List, Optional, Union

from pydash import _

from ..bootstrap import _CLI


DEFAULTS = _CLI().get_config()

class BaseRequest:
    """Assertio Request object."""

    def __init__(self, method: Optional[str] = None):
        """Class constructor."""
        self._body: Union[Dict, None] = None
        self._endpoint: Union[str, None] = None
        self._headers: Union[Dict, None] = None
        self._method: Union[str, None] = method
        self._params: Union[Dict, None] = None

    @property
    def body(self):
        """Body getter."""
        if isinstance(self._body, str):
            return json.loads(self._body)
        return self._body
    
    @body.setter
    def body(self, body: Union[Dict, str]):
        """Body setter."""
        if isinstance(body, str):
            body = Path.cwd().joinpath(f"{DEFAULTS.payloads_dir}/{body}")
        if isinstance(body, Path):
            with open(body) as stream:
                body = json.load(stream)

        self._body = json.dumps(body)

    @property
    def endpoint(self):
        """Endpoint getter, no setter available."""
        return self._endpoint

    @endpoint.setter
    def endpoint(self, new_endpoint):
        """Endpoint setter."""
        self._endpoint = new_endpoint

    @property
    def headers(self):
        """Headers getter."""
        return self._headers
    
    @headers.setter
    def headers(self, new_headers):
        """Headers setter."""
        if self._headers is None:
            self._headers = new_headers
        else:
            self._headers.update(new_headers)

    @property
    def method(self):
        """Method getter."""
        return self._method

    @method.setter
    def method(self, new_method):
        """Method setter."""
        self._method = new_method

    @property
    def params(self):
        return self._params
    
    @params.setter
    def params(self, new_params):
        if self._params is None:
            self._params = new_params
        else:
            self._params.update(new_params)

    def get_response_field(self, path: Union[str, List[str]]):
        """Return value from response payload.
        Must be used after perform()
        
        """
        return _.get(self.response.json(), path)


"""HTTP status module."""
from http import HTTPStatus as HTTP
from typing import Optional

from truth.truth import AssertThat

from ...decorators import then
from ..base_request import BaseRequest


class StatusAssertions(BaseRequest):
    """HTTP status assertions."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    @then
    def assert_http_continue(self):
        """Assert response status is 100 CONTINUE."""
        AssertThat(self.response.status_code).IsEqualTo(HTTP.CONTINUE)

    @then
    def assert_http_switching_protocols(self):
        """Assert response status is 101 SWITCHING_PROTOCOLS."""
        AssertThat(self.response.status_code).IsEqualTo(
            HTTP.SWITCHING_PROTOCOLS
        )

    @then
    def assert_http_processing(self):
        """Assert response status is 102 PROCESSING."""
        AssertThat(self.response.status_code).IsEqualTo(HTTP.PROCESSING)

    @then
    def assert_http_early_hints(self):
        """Assert response status is 103 EARLY_HINTS."""
        AssertThat(self.response.status_code).IsEqualTo(HTTP.EARLY_HINTS)

    @then
    def assert_http_ok(self):
        """Assert response status is 200 OK."""
        AssertThat(self.response.status_code).IsEqualTo(HTTP.OK)

    @then
    def assert_http_created(self):
        """Assert response status is 201 CREATED."""
        AssertThat(self.response.status_code).IsEqualTo(HTTP.CREATED)

    @then
    def assert_http_accepted(self):
        """Assert response status is 202 ACCEPTED."""
        AssertThat(self.response.status_code).IsEqualTo(HTTP.ACCEPTED)

    @then
    def assert_http_non_authoritative_information(self):
        """Assert response status is 203 NON_AUTHORITATIVE_INFORMATION."""
        AssertThat(self.response.status_code).IsEqualTo(
            HTTP.NON_AUTHORITATIVE_INFORMATION
        )

    @then
    def assert_http_no_content(self):
        """Assert response status is 204 NO_CONTENT."""
        AssertThat(self.response.status_code).IsEqualTo(HTTP.NO_CONTENT)

    @then
    def assert_http_reset_content(self):
        """Assert response status is 205 RESET_CONTENT."""
        AssertThat(self.response.status_code).IsEqualTo(HTTP.RESET_CONTENT)

    @then
    def assert_http_partial_content(self):
        """Assert response status is 206 PARTIAL_CONTENT."""
        AssertThat(self.response.status_code).IsEqualTo(HTTP.PARTIAL_CONTENT)

    @then
    def assert_http_multi_status(self):
        """Assert response status is 207 MULTI_STATUS."""
        AssertThat(self.response.status_code).IsEqualTo(HTTP.MULTI_STATUS)

    @then
    def assert_http_already_reported(self):
        """Assert response status is 208 ALREADY_REPORTED."""
        AssertThat(self.response.status_code).IsEqualTo(HTTP.ALREADY_REPORTED)

    @then
    def assert_http_im_used(self):
        """Assert response status is 226 IM_USED."""
        AssertThat(self.response.status_code).IsEqualTo(HTTP.IM_USED)

    @then
    def assert_http_multiple_choices(self):
        """Assert response status is 300 MULTIPLE_CHOICES."""
        AssertThat(self.response.status_code).IsEqualTo(HTTP.MULTIPLE_CHOICES)

    @then
    def assert_http_moved_permanently(self):
        """Assert response status is 301 MOVED_PERMANENTLY."""
        AssertThat(self.response.status_code).IsEqualTo(HTTP.MOVED_PERMANENTLY)

    @then
    def assert_http_found(self):
        """Assert response status is 302 FOUND."""
        AssertThat(self.response.status_code).IsEqualTo(HTTP.FOUND)

    @then
    def assert_http_see_other(self):
        """Assert response status is 303 SEE_OTHER."""
        AssertThat(self.response.status_code).IsEqualTo(HTTP.SEE_OTHER)

    @then
    def assert_http_not_modified(self):
        """Assert response status is 304 NOT_MODIFIED."""
        AssertThat(self.response.status_code).IsEqualTo(HTTP.NOT_MODIFIED)

    @then
    def assert_http_use_proxy(self):
        """Assert response status is 305 USE_PROXY."""
        AssertThat(self.response.status_code).IsEqualTo(HTTP.USE_PROXY)

    @then
    def assert_http_temporary_redirect(self):
        """Assert response status is 307 TEMPORARY_REDIRECT."""
        AssertThat(self.response.status_code).IsEqualTo(
            HTTP.TEMPORARY_REDIRECT
        )

    @then
    def assert_http_permanent_redirect(self):
        """Assert response status is 308 PERMANENT_REDIRECT."""
        AssertThat(self.response.status_code).IsEqualTo(
            HTTP.PERMANENT_REDIRECT
        )

    @then
    def assert_http_bad_request(self):
        """Assert response status is 400 BAD_REQUEST."""
        AssertThat(self.response.status_code).IsEqualTo(HTTP.BAD_REQUEST)

    @then
    def assert_http_unauthorized(self):
        """Assert response status is 401 UNAUTHORIZED."""
        AssertThat(self.response.status_code).IsEqualTo(HTTP.UNAUTHORIZED)

    @then
    def assert_http_payment_required(self):
        """Assert response status is 402 PAYMENT_REQUIRED."""
        AssertThat(self.response.status_code).IsEqualTo(HTTP.PAYMENT_REQUIRED)

    @then
    def assert_http_forbidden(self):
        """Assert response status is 403 FORBIDDEN."""
        AssertThat(self.response.status_code).IsEqualTo(HTTP.FORBIDDEN)

    @then
    def assert_http_not_found(self):
        """Assert response status is 404 NOT_FOUND."""
        AssertThat(self.response.status_code).IsEqualTo(HTTP.NOT_FOUND)

    @then
    def assert_http_method_not_allowed(self):
        """Assert response status is 405 METHOD_NOT_ALLOWED."""
        AssertThat(self.response.status_code).IsEqualTo(
            HTTP.METHOD_NOT_ALLOWED
        )

    @then
    def assert_http_not_acceptable(self):
        """Assert response status is 406 NOT_ACCEPTABLE."""
        AssertThat(self.response.status_code).IsEqualTo(HTTP.NOT_ACCEPTABLE)

    @then
    def assert_http_proxy_authentication_required(self):
        """Assert response status is 407 PROXY_AUTHENTICATION_REQUIRED."""
        AssertThat(self.response.status_code).IsEqualTo(
            HTTP.PROXY_AUTHENTICATION_REQUIRED
        )

    @then
    def assert_http_request_timeout(self):
        """Assert response status is 408 REQUEST_TIMEOUT."""
        AssertThat(self.response.status_code).IsEqualTo(HTTP.REQUEST_TIMEOUT)

    @then
    def assert_http_conflict(self):
        """Assert response status is 409 CONFLICT."""
        AssertThat(self.response.status_code).IsEqualTo(HTTP.CONFLICT)

    @then
    def assert_http_gone(self):
        """Assert response status is 410 GONE."""
        AssertThat(self.response.status_code).IsEqualTo(HTTP.GONE)

    @then
    def assert_http_length_required(self):
        """Assert response status is 411 LENGTH_REQUIRED."""
        AssertThat(self.response.status_code).IsEqualTo(HTTP.LENGTH_REQUIRED)

    @then
    def assert_http_precondition_failed(self):
        """Assert response status is 412 PRECONDITION_FAILED."""
        AssertThat(self.response.status_code).IsEqualTo(
            HTTP.PRECONDITION_FAILED
        )

    @then
    def assert_http_request_entity_too_large(self):
        """Assert response status is 413 REQUEST_ENTITY_TOO_LARGE."""
        AssertThat(self.response.status_code).IsEqualTo(
            HTTP.REQUEST_ENTITY_TOO_LARGE
        )

    @then
    def assert_http_request_uri_too_long(self):
        """Assert response status is 414 REQUEST_URI_TOO_LONG."""
        AssertThat(self.response.status_code).IsEqualTo(
            HTTP.REQUEST_URI_TOO_LONG
        )

    @then
    def assert_http_unsupported_media_type(self):
        """Assert response status is 415 UNSUPPORTED_MEDIA_TYPE."""
        AssertThat(self.response.status_code).IsEqualTo(
            HTTP.UNSUPPORTED_MEDIA_TYPE
        )

    @then
    def assert_http_requested_range_not_satisfiable(self):
        """Assert response status is 416 REQUESTED_RANGE_NOT_SATISFIABLE."""
        AssertThat(self.response.status_code).IsEqualTo(
            HTTP.REQUESTED_RANGE_NOT_SATISFIABLE
        )

    @then
    def assert_http_expectation_failed(self):
        """Assert response status is 417 EXPECTATION_FAILED."""
        AssertThat(self.response.status_code).IsEqualTo(
            HTTP.EXPECTATION_FAILED
        )

    @then
    def assert_http_im_a_teapot(self):
        """Assert response status is 418 IM_A_TEAPOT."""
        AssertThat(self.response.status_code).IsEqualTo(HTTP.IM_A_TEAPOT)

    @then
    def assert_http_misdirected_request(self):
        """Assert response status is 421 MISDIRECTED_REQUEST."""
        AssertThat(self.response.status_code).IsEqualTo(
            HTTP.MISDIRECTED_REQUEST
        )

    @then
    def assert_http_unprocessable_entity(self):
        """Assert response status is 422 UNPROCESSABLE_ENTITY."""
        AssertThat(self.response.status_code).IsEqualTo(
            HTTP.UNPROCESSABLE_ENTITY
        )

    @then
    def assert_http_locked(self):
        """Assert response status is 423 LOCKED."""
        AssertThat(self.response.status_code).IsEqualTo(HTTP.LOCKED)

    @then
    def assert_http_failed_dependency(self):
        """Assert response status is 424 FAILED_DEPENDENCY."""
        AssertThat(self.response.status_code).IsEqualTo(HTTP.FAILED_DEPENDENCY)

    @then
    def assert_http_too_early(self):
        """Assert response status is 425 TOO_EARLY."""
        AssertThat(self.response.status_code).IsEqualTo(HTTP.TOO_EARLY)

    @then
    def assert_http_upgrade_required(self):
        """Assert response status is 426 UPGRADE_REQUIRED."""
        AssertThat(self.response.status_code).IsEqualTo(HTTP.UPGRADE_REQUIRED)

    @then
    def assert_http_precondition_required(self):
        """Assert response status is 428 PRECONDITION_REQUIRED."""
        AssertThat(self.response.status_code).IsEqualTo(
            HTTP.PRECONDITION_REQUIRED
        )

    @then
    def assert_http_too_many_requests(self):
        """Assert response status is 429 TOO_MANY_REQUESTS."""
        AssertThat(self.response.status_code).IsEqualTo(HTTP.TOO_MANY_REQUESTS)

    @then
    def assert_http_request_header_fields_too_large(self):
        """Assert response status is 431 REQUEST_HEADER_FIELDS_TOO_LARGE."""
        AssertThat(self.response.status_code).IsEqualTo(
            HTTP.REQUEST_HEADER_FIELDS_TOO_LARGE
        )

    @then
    def assert_http_unavailable_for_legal_reasons(self):
        """Assert response status is 451 UNAVAILABLE_FOR_LEGAL_REASONS."""
        AssertThat(self.response.status_code).IsEqualTo(
            HTTP.UNAVAILABLE_FOR_LEGAL_REASONS
        )

    @then
    def assert_http_internal_server_error(self):
        """Assert response status is 500 INTERNAL_SERVER_ERROR."""
        AssertThat(self.response.status_code).IsEqualTo(
            HTTP.INTERNAL_SERVER_ERROR
        )

    @then
    def assert_http_not_implemented(self):
        """Assert response status is 501 NOT_IMPLEMENTED."""
        AssertThat(self.response.status_code).IsEqualTo(HTTP.NOT_IMPLEMENTED)

    @then
    def assert_http_bad_gateway(self):
        """Assert response status is 502 BAD_GATEWAY."""
        AssertThat(self.response.status_code).IsEqualTo(HTTP.BAD_GATEWAY)

    @then
    def assert_http_service_unavailable(self):
        """Assert response status is 503 SERVICE_UNAVAILABLE."""
        AssertThat(self.response.status_code).IsEqualTo(
            HTTP.SERVICE_UNAVAILABLE
        )

    @then
    def assert_http_gateway_timeout(self):
        """Assert response status is 504 GATEWAY_TIMEOUT."""
        AssertThat(self.response.status_code).IsEqualTo(HTTP.GATEWAY_TIMEOUT)

    @then
    def assert_http_http_version_not_supported(self):
        """Assert response status is 505 HTTP_VERSION_NOT_SUPPORTED."""
        AssertThat(self.response.status_code).IsEqualTo(
            HTTP.HTTP_VERSION_NOT_SUPPORTED
        )

    @then
    def assert_http_variant_also_negotiates(self):
        """Assert response status is 506 VARIANT_ALSO_NEGOTIATES."""
        AssertThat(self.response.status_code).IsEqualTo(
            HTTP.VARIANT_ALSO_NEGOTIATES
        )

    @then
    def assert_http_insufficient_storage(self):
        """Assert response status is 507 INSUFFICIENT_STORAGE."""
        AssertThat(self.response.status_code).IsEqualTo(
            HTTP.INSUFFICIENT_STORAGE
        )

    @then
    def assert_http_loop_detected(self):
        """Assert response status is 508 LOOP_DETECTED."""
        AssertThat(self.response.status_code).IsEqualTo(HTTP.LOOP_DETECTED)

    @then
    def assert_http_not_extended(self):
        """Assert response status is 510 NOT_EXTENDED."""
        AssertThat(self.response.status_code).IsEqualTo(HTTP.NOT_EXTENDED)

    @then
    def assert_http_network_authentication_required(self):
        """Assert response status is 511 NETWORK_AUTHENTICATION_REQUIRED."""
        AssertThat(self.response.status_code).IsEqualTo(
            HTTP.NETWORK_AUTHENTICATION_REQUIRED
        )



"""Decorators module."""
from loguru import logger

from ..bootstrap import _CLI


DEFAULTS = _CLI().get_config()
logger.add(DEFAULTS.logfile, format="{time} | {level} | {message}")


def given(fn):
    """Set a function as a given precondition."""

    def _(instance, *args, **kwargs):
        fn(instance, *args, **kwargs)
        return instance

    return _


def then(fn):
    """Set a function as a then action."""

    def _(instance, *args, **kwargs):
        msg = (
            f"{instance.method.upper()} {DEFAULTS.base_url}{instance.endpoint}"
        )
        try:
            fn(instance, *args, **kwargs)
            logger.success(f"PASSED {msg} {fn.__name__}")
        except AssertionError:
            logger.error(f"FAILED: {msg} {fn.__name__}")
        finally:
            return instance

    return _


def when(fn):
    """Set a function as a when assertion."""

    def _(instance, *args, **kwargs):
        fn(instance, *args, **kwargs)
        return instance

    return _


def log_test(fn):
    """Add a simple log message when a test starts."""
    def _(*args, **kwargs):
        logger.info(f"{fn.__name__} started")
        fn(*args, **kwargs)

    return _



"""Aliases and exports."""
from .decorators import given, then, when, log_test


Given, Then, When = given, then, when
g, t, w = given, then, when


__all__ = ("Given", "Then", "When", "given", "then", "when", "g", "t", "w")



"""Test runner module."""
from .decorators import log_test


class Runner:
    """Test runner."""

    def start(self, *args):
        """Run all tests on a runner.

        Tests function names must start with 'test'.
        """
        for fn in filter(lambda fn: fn.startswith("test"), dir(self)):
            log_test(getattr(self, fn))(*args)



"""Preconditions module."""
from typing import Dict, Union

from ..decorators import given
from .base_request import BaseRequest


class Preconditions(BaseRequest):
    """Precondition methods."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    @given
    def to(self, endpoint, **kwargs):
        """Set endpoint to request."""
        self.endpoint = endpoint
        if kwargs:
            self.endpoint = self.endpoint.format(**kwargs)

    @given
    def with_method(self, method):
        """Set HTTP request method."""
        self.method = method

    @given
    def with_body(self, body: Union[Dict, str]):
        """Set request Content-Type: appliaction/json body."""
        self.body = body

    @given
    def with_headers(self, headers):
        """Set request header or headers."""
        self.headers = headers

    @given
    def with_params(self, params):
        """Set request query parameters."""
        self.params = params



"""Assertio request module."""
from json import dumps
from requests import request

from ..decorators import when
from ..bootstrap import _CLI
from .base_request import BaseRequest


DEFAULTS = _CLI().get_config()

class Actions(BaseRequest):
    """Assertio Request object."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    @when
    def perform(self):
        """Execute request."""
        self.response = request(
            self.method,
            f"{DEFAULTS.base_url}{self.endpoint}",
            params=self.params,
            data=dumps(self.body),
            headers=self.headers,
        )
